@@ -21,6 +21,7 @@
 
 #include "gtkinscriptionprivate.h"
 
+#include "gtkaccessibletextprivate.h"
 #include "gtkcssnodeprivate.h"
 #include "gtkcssstylechangeprivate.h"
 #include "gtkpangoprivate.h"
@@ -29,6 +30,8 @@
 #include "gtktypebuiltins.h"
 #include "gtkwidgetprivate.h"
 
+#include "a11y/gtkatspipangoprivate.h"
+
 /**
  * GtkInscription:
  *
@@ -96,7 +99,11 @@ enum
   N_PROPS
 };
 
-G_DEFINE_TYPE (GtkInscription, gtk_inscription, GTK_TYPE_WIDGET)
+static void     gtk_inscription_accessible_text_init (GtkAccessibleTextInterface *iface);
+
+G_DEFINE_TYPE_WITH_CODE (GtkInscription, gtk_inscription, GTK_TYPE_WIDGET,
+                         G_IMPLEMENT_INTERFACE (GTK_TYPE_ACCESSIBLE_TEXT,
+                                                gtk_inscription_accessible_text_init))
 
 static GParamSpec *properties[N_PROPS] = { NULL, };
 
@@ -454,6 +461,40 @@ gtk_inscription_get_layout_location (GtkInscription *self,
   *y_out = y;
 }
 
+static gboolean
+gtk_inscription_get_layout_index (GtkInscription *self,
+                                  int             x,
+                                  int             y,
+                                  int            *index)
+{
+  int trailing = 0;
+  const char *cluster;
+  const char *cluster_end;
+  gboolean inside;
+  float lx, ly;
+
+  *index = 0;
+
+  gtk_inscription_get_layout_location (self, &lx, &ly);
+
+  inside = pango_layout_xy_to_index (self->layout,
+                                     (x - lx) * PANGO_SCALE,
+                                     (y - ly) * PANGO_SCALE,
+                                     index, &trailing);
+
+  cluster = self->text + *index;
+  cluster_end = cluster;
+  while (trailing)
+    {
+      cluster_end = g_utf8_next_char (cluster_end);
+      --trailing;
+    }
+
+  *index += (cluster_end - cluster);
+
+  return inside;
+}
+
 static void
 gtk_inscription_allocate (GtkWidget *widget,
                           int        width,
@@ -544,7 +585,7 @@ gtk_inscription_class_init (GtkInscriptionClass *klass)
   widget_class->snapshot = gtk_inscription_snapshot;
 
   /**
-   * GtkInscription:attributes: (attributes org.gtk.Property.get=gtk_inscription_get_attributes org.gtk.Property.set=gtk_inscription_set_attributes)
+   * GtkInscription:attributes:
    *
    * A list of style attributes to apply to the text of the inscription.
    *
@@ -556,7 +597,7 @@ gtk_inscription_class_init (GtkInscriptionClass *klass)
                           G_PARAM_READWRITE | G_PARAM_EXPLICIT_NOTIFY | G_PARAM_STATIC_STRINGS);
 
   /**
-   * GtkInscription:markup: (attributes org.gtk.Property.set=gtk_inscription_set_markup)
+   * GtkInscription:markup:
    *
    * Utility property that sets both the [property@Gtk.Inscription:text] and
    * [property@Gtk.Inscription:attributes] properties, mainly intended for use in
@@ -574,7 +615,7 @@ gtk_inscription_class_init (GtkInscriptionClass *klass)
                          G_PARAM_WRITABLE | G_PARAM_EXPLICIT_NOTIFY | G_PARAM_STATIC_STRINGS);
 
   /**
-   * GtkInscription:min-chars: (attributes org.gtk.Property.get=gtk_inscription_get_min_chars org.gtk.Property.set=gtk_inscription_set_min_chars)
+   * GtkInscription:min-chars:
    *
    * The number of characters that should fit into the inscription at minimum.
    *
@@ -597,7 +638,7 @@ gtk_inscription_class_init (GtkInscriptionClass *klass)
                          G_PARAM_READWRITE | G_PARAM_EXPLICIT_NOTIFY | G_PARAM_STATIC_STRINGS);
 
   /**
-   * GtkInscription:min-lines: (attributes org.gtk.Property.get=gtk_inscription_get_min_lines org.gtk.Property.set=gtk_inscription_set_min_lines)
+   * GtkInscription:min-lines:
    *
    * The number of lines that should fit into the inscription at minimum.
    *
@@ -619,7 +660,7 @@ gtk_inscription_class_init (GtkInscriptionClass *klass)
                          G_PARAM_READWRITE | G_PARAM_EXPLICIT_NOTIFY | G_PARAM_STATIC_STRINGS);
 
   /**
-   * GtkInscription:nat-chars: (attributes org.gtk.Property.get=gtk_inscription_get_nat_chars org.gtk.Property.set=gtk_inscription_set_nat_chars)
+   * GtkInscription:nat-chars:
    *
    * The number of characters that should ideally fit into the inscription.
    *
@@ -639,7 +680,7 @@ gtk_inscription_class_init (GtkInscriptionClass *klass)
                          G_PARAM_READWRITE | G_PARAM_EXPLICIT_NOTIFY | G_PARAM_STATIC_STRINGS);
 
   /**
-   * GtkInscription:nat-lines: (attributes org.gtk.Property.get=gtk_inscription_get_nat_lines org.gtk.Property.set=gtk_inscription_set_nat_lines)
+   * GtkInscription:nat-lines:
    *
    * The number of lines that should ideally fit into the inscription.
    *
@@ -659,7 +700,7 @@ gtk_inscription_class_init (GtkInscriptionClass *klass)
                          G_PARAM_READWRITE | G_PARAM_EXPLICIT_NOTIFY | G_PARAM_STATIC_STRINGS);
 
   /**
-   * GtkInscription:text: (attributes org.gtk.Property.get=gtk_inscription_get_text org.gtk.Property.set=gtk_inscription_set_text)
+   * GtkInscription:text:
    *
    * The displayed text.
    *
@@ -671,7 +712,7 @@ gtk_inscription_class_init (GtkInscriptionClass *klass)
                          G_PARAM_READWRITE | G_PARAM_EXPLICIT_NOTIFY | G_PARAM_STATIC_STRINGS);
 
   /**
-   * GtkInscription:text-overflow: (attributes org.gtk.Property.get=gtk_inscription_get_text_overflow org.gtk.Property.set=gtk_inscription_set_text_overflow)
+   * GtkInscription:text-overflow:
    *
    * The overflow method to use for the text.
    *
@@ -684,7 +725,7 @@ gtk_inscription_class_init (GtkInscriptionClass *klass)
                        G_PARAM_READWRITE | G_PARAM_EXPLICIT_NOTIFY | G_PARAM_STATIC_STRINGS);
 
   /**
-   * GtkInscription:wrap-mode: (attributes org.gtk.Property.get=gtk_inscription_get_wrap_mode org.gtk.Property.set=gtk_inscription_set_wrap_mode)
+   * GtkInscription:wrap-mode:
    *
    * Controls how the line wrapping is done.
    *
@@ -699,7 +740,7 @@ gtk_inscription_class_init (GtkInscriptionClass *klass)
                          G_PARAM_READWRITE | G_PARAM_EXPLICIT_NOTIFY | G_PARAM_STATIC_STRINGS);
 
   /**
-   * GtkInscription:xalign: (attributes org.gtk.Property.get=gtk_inscription_get_xalign org.gtk.Property.set=gtk_inscription_set_xalign)
+   * GtkInscription:xalign:
    *
    * The horizontal alignment of the text inside the allocated size.
    *
@@ -715,7 +756,7 @@ gtk_inscription_class_init (GtkInscriptionClass *klass)
                           G_PARAM_READWRITE | G_PARAM_EXPLICIT_NOTIFY | G_PARAM_STATIC_STRINGS);
 
   /**
-   * GtkInscription:yalign: (attributes org.gtk.Property.get=gtk_inscription_get_yalign org.gtk.Property.set=gtk_inscription_set_yalign)
+   * GtkInscription:yalign:
    *
    * The vertical alignment of the text inside the allocated size.
    *
@@ -777,7 +818,7 @@ gtk_inscription_new (const char *text)
 }
 
 /**
- * gtk_inscription_set_text: (attributes org.gtk.Method.set_property=text)
+ * gtk_inscription_set_text:
  * @self: a `GtkInscription`
  * @text: (nullable): The text to display
  *
@@ -805,10 +846,15 @@ gtk_inscription_set_text (GtkInscription *self,
   gtk_widget_queue_draw (GTK_WIDGET (self));
 
   g_object_notify_by_pspec (G_OBJECT (self), properties[PROP_TEXT]);
+
+  gtk_accessible_update_property (GTK_ACCESSIBLE (self),
+                                  GTK_ACCESSIBLE_PROPERTY_LABEL,
+                                  text,
+                                  -1);
 }
 
 /**
- * gtk_inscription_get_text: (attributes org.gtk.Method.get_property=text)
+ * gtk_inscription_get_text:
  * @self: a `GtkInscription`
  *
  * Gets the text that is displayed.
@@ -826,7 +872,7 @@ gtk_inscription_get_text (GtkInscription *self)
 }
 
 /**
- * gtk_inscription_set_min_chars: (attributes org.gtk.Method.set_property=min-chars)
+ * gtk_inscription_set_min_chars:
  * @self: a `GtkInscription`
  * @min_chars: the minimum number of characters that should fit, approximately
  *
@@ -853,7 +899,7 @@ gtk_inscription_set_min_chars (GtkInscription *self,
 }
 
 /**
- * gtk_inscription_get_min_chars: (attributes org.gtk.Method.get_property=min-chars)
+ * gtk_inscription_get_min_chars:
  * @self: a `GtkInscription`
  *
  * Gets the `min-chars` of the inscription.
@@ -873,7 +919,7 @@ gtk_inscription_get_min_chars (GtkInscription *self)
 }
 
 /**
- * gtk_inscription_set_nat_chars: (attributes org.gtk.Method.set_property=nat-chars)
+ * gtk_inscription_set_nat_chars:
  * @self: a `GtkInscription`
  * @nat_chars: the number of characters that should ideally fit, approximately
  *
@@ -900,7 +946,7 @@ gtk_inscription_set_nat_chars (GtkInscription *self,
 }
 
 /**
- * gtk_inscription_get_nat_chars: (attributes org.gtk.Method.get_property=nat-chars)
+ * gtk_inscription_get_nat_chars:
  * @self: a `GtkInscription`
  *
  * Gets the `nat-chars` of the inscription.
@@ -920,7 +966,7 @@ gtk_inscription_get_nat_chars (GtkInscription *self)
 }
 
 /**
- * gtk_inscription_set_min_lines: (attributes org.gtk.Method.set_property=min-lines)
+ * gtk_inscription_set_min_lines:
  * @self: a `GtkInscription`
  * @min_lines: the minimum number of lines that should fit, approximately
  *
@@ -947,7 +993,7 @@ gtk_inscription_set_min_lines (GtkInscription *self,
 }
 
 /**
- * gtk_inscription_get_min_lines: (attributes org.gtk.Method.get_property=min-lines)
+ * gtk_inscription_get_min_lines:
  * @self: a `GtkInscription`
  *
  * Gets the `min-lines` of the inscription.
@@ -967,7 +1013,7 @@ gtk_inscription_get_min_lines (GtkInscription *self)
 }
 
 /**
- * gtk_inscription_set_nat_lines: (attributes org.gtk.Method.set_property=nat-lines)
+ * gtk_inscription_set_nat_lines:
  * @self: a `GtkInscription`
  * @nat_lines: the number of lines that should ideally fit
  *
@@ -994,7 +1040,7 @@ gtk_inscription_set_nat_lines (GtkInscription *self,
 }
 
 /**
- * gtk_inscription_get_nat_lines: (attributes org.gtk.Method.get_property=nat-lines)
+ * gtk_inscription_get_nat_lines:
  * @self: a `GtkInscription`
  *
  * Gets the `nat-lines` of the inscription.
@@ -1014,7 +1060,7 @@ gtk_inscription_get_nat_lines (GtkInscription *self)
 }
 
 /**
- * gtk_inscription_set_xalign: (attributes org.gtk.Method.set_property=xalign)
+ * gtk_inscription_set_xalign:
  * @self: a `GtkInscription`
  * @xalign: the new xalign value, between 0 and 1
  *
@@ -1045,7 +1091,7 @@ gtk_inscription_set_xalign (GtkInscription *self,
 }
 
 /**
- * gtk_inscription_get_xalign: (attributes org.gtk.Method.get_property=xalign)
+ * gtk_inscription_get_xalign:
  * @self: a `GtkInscription`
  *
  * Gets the `xalign` of the inscription.
@@ -1065,7 +1111,7 @@ gtk_inscription_get_xalign (GtkInscription *self)
 }
 
 /**
- * gtk_inscription_set_yalign: (attributes org.gtk.Method.set_property=yalign)
+ * gtk_inscription_set_yalign:
  * @self: a `GtkInscription`
  * @yalign: the new yalign value, between 0 and 1
  *
@@ -1094,7 +1140,7 @@ gtk_inscription_set_yalign (GtkInscription *self,
 }
 
 /**
- * gtk_inscription_get_yalign: (attributes org.gtk.Method.get_property=yalign)
+ * gtk_inscription_get_yalign:
  * @self: a `GtkInscription`
  *
  * Gets the `yalign` of the inscription.
@@ -1114,7 +1160,7 @@ gtk_inscription_get_yalign (GtkInscription *self)
 }
 
 /**
- * gtk_inscription_set_attributes: (attributes org.gtk.Method.set_property=attributes)
+ * gtk_inscription_set_attributes:
  * @self: a `GtkInscription`
  * @attrs: (nullable): a [struct@Pango.AttrList]
  *
@@ -1148,7 +1194,7 @@ gtk_inscription_set_attributes (GtkInscription *self,
 }
 
 /**
- * gtk_inscription_get_attributes: (attributes org.gtk.Method.get_property=attributes)
+ * gtk_inscription_get_attributes:
  * @self: a `GtkInscription`
  *
  * Gets the inscription's attribute list.
@@ -1166,7 +1212,7 @@ gtk_inscription_get_attributes (GtkInscription *self)
 }
 
 /**
- * gtk_inscription_set_text_overflow: (attributes org.gtk.Method.set_property=text-overflow)
+ * gtk_inscription_set_text_overflow:
  * @self: a `GtkInscription`
  * @overflow: the overflow method to use
  *
@@ -1210,7 +1256,7 @@ gtk_inscription_set_text_overflow (GtkInscription         *self,
 }
 
 /**
- * gtk_inscription_get_text_overflow: (attributes org.gtk.Method.get_property=text-overflow)
+ * gtk_inscription_get_text_overflow:
  * @self: a `GtkInscription`
  *
  * Gets the inscription's overflow method.
@@ -1228,7 +1274,7 @@ gtk_inscription_get_text_overflow (GtkInscription *self)
 }
 
 /**
- * gtk_inscription_set_wrap_mode: (attributes org.gtk.Method.set_property=wrap-mode)
+ * gtk_inscription_set_wrap_mode:
  * @self: a `GtkInscription`
  * @wrap_mode: the line wrapping mode
  *
@@ -1253,7 +1299,7 @@ gtk_inscription_set_wrap_mode (GtkInscription *self,
 }
 
 /**
- * gtk_inscription_get_wrap_mode: (attributes org.gtk.Method.get_property=wrap-mode)
+ * gtk_inscription_get_wrap_mode:
  * @self: a `GtkInscription`
  *
  * Returns line wrap mode used by the inscription.
@@ -1273,7 +1319,7 @@ gtk_inscription_get_wrap_mode (GtkInscription *self)
 }
 
 /**
- * gtk_inscription_set_markup: (attributes org.gtk.Method.set_property=markup)
+ * gtk_inscription_set_markup:
  * @self: a `GtkInscription`
  * @markup: (nullable): The markup to display
  *
@@ -1315,3 +1361,176 @@ gtk_inscription_set_markup (GtkInscription *self,
   g_clear_pointer (&text, g_free);
   g_clear_pointer (&attrs, pango_attr_list_unref);
 }
+
+/* {{{ GtkAccessibleText implementation */
+
+static GBytes *
+gtk_inscription_accessible_text_get_contents (GtkAccessibleText *self,
+                                              unsigned int       start,
+                                              unsigned int       end)
+{
+  const char *text;
+  int len;
+  char *string;
+  gsize size;
+
+  text = gtk_inscription_get_text (GTK_INSCRIPTION (self));
+  len = g_utf8_strlen (text, -1);
+
+  start = CLAMP (start, 0, len);
+  end = CLAMP (end, 0, len);
+
+  if (end <= start)
+    {
+      string = g_strdup ("");
+      size = 1;
+    }
+  else
+    {
+      const char *p, *q;
+      p = g_utf8_offset_to_pointer (text, start);
+      q = g_utf8_offset_to_pointer (text, end);
+      size = q - p + 1;
+      string = g_strndup (p, q - p);
+    }
+
+  return g_bytes_new_take (string, size);
+}
+
+static GBytes *
+gtk_inscription_accessible_text_get_contents_at (GtkAccessibleText            *self,
+                                                 unsigned int                  offset,
+                                                 GtkAccessibleTextGranularity  granularity,
+                                                 unsigned int                 *start,
+                                                 unsigned int                 *end)
+{
+  PangoLayout *layout = gtk_inscription_get_layout (GTK_INSCRIPTION (self));
+  char *string = gtk_pango_get_string_at (layout, offset, granularity, start, end);
+
+  return g_bytes_new_take (string, strlen (string));
+}
+
+static unsigned int
+gtk_inscription_accessible_text_get_caret_position (GtkAccessibleText *self)
+{
+  return 0;
+}
+
+static gboolean
+gtk_inscription_accessible_text_get_selection (GtkAccessibleText       *self,
+                                               gsize                   *n_ranges,
+                                               GtkAccessibleTextRange **ranges)
+{
+  return FALSE;
+}
+
+static gboolean
+gtk_inscription_accessible_text_get_attributes (GtkAccessibleText        *self,
+                                                unsigned int              offset,
+                                                gsize                    *n_ranges,
+                                                GtkAccessibleTextRange  **ranges,
+                                                char                   ***attribute_names,
+                                                char                   ***attribute_values)
+{
+  PangoLayout *layout = gtk_inscription_get_layout (GTK_INSCRIPTION (self));
+  unsigned int start, end;
+  char **names, **values;
+
+  gtk_pango_get_run_attributes (layout, offset, &names, &values, &start, &end);
+
+  *n_ranges = g_strv_length (names);
+  *ranges = g_new (GtkAccessibleTextRange, *n_ranges);
+
+  for (unsigned i = 0; i < *n_ranges; i++)
+    {
+      GtkAccessibleTextRange *range = &(*ranges)[i];
+
+      range->start = start;
+      range->length = end - start;
+    }
+
+  *attribute_names = names;
+  *attribute_values = values;
+
+  return TRUE;
+}
+
+static void
+gtk_inscription_accessible_text_get_default_attributes (GtkAccessibleText   *self,
+                                                        char              ***attribute_names,
+                                                        char              ***attribute_values)
+{
+  PangoLayout *layout = gtk_inscription_get_layout (GTK_INSCRIPTION (self));
+  char **names, **values;
+
+  gtk_pango_get_default_attributes (layout, &names, &values);
+
+  *attribute_names = names;
+  *attribute_values = values;
+}
+
+static gboolean
+gtk_inscription_accessible_text_get_extents (GtkAccessibleText *self,
+                                             unsigned int       start,
+                                             unsigned int       end,
+                                             graphene_rect_t   *extents)
+{
+  GtkInscription *inscription = GTK_INSCRIPTION (self);
+  PangoLayout *layout;
+  const char *text;
+  float lx, ly;
+  cairo_region_t *range_clip;
+  cairo_rectangle_int_t clip_rect;
+  int range[2];
+
+  layout = inscription->layout;
+  text = inscription->text;
+  gtk_inscription_get_layout_location (inscription, &lx, &ly);
+
+  range[0] = g_utf8_pointer_to_offset (text, text + start);
+  range[1] = g_utf8_pointer_to_offset (text, text + end);
+
+  range_clip = gdk_pango_layout_get_clip_region (layout, lx, ly, range, 1);
+  cairo_region_get_extents (range_clip, &clip_rect);
+  cairo_region_destroy (range_clip);
+
+  extents->origin.x = clip_rect.x;
+  extents->origin.y = clip_rect.y;
+  extents->size.width = clip_rect.width;
+  extents->size.height = clip_rect.height;
+
+  return TRUE;
+}
+
+static gboolean
+gtk_inscription_accessible_text_get_offset (GtkAccessibleText      *self,
+                                            const graphene_point_t *point,
+                                            unsigned int           *offset)
+{
+  GtkInscription *inscription = GTK_INSCRIPTION (self);
+  int index;
+
+  if (!gtk_inscription_get_layout_index (inscription, point->x, point->y, &index))
+    return FALSE;
+
+  *offset = (unsigned int) g_utf8_pointer_to_offset (inscription->text, inscription->text + index);
+
+  return TRUE;
+}
+
+static void
+gtk_inscription_accessible_text_init (GtkAccessibleTextInterface *iface)
+{
+  iface->get_contents = gtk_inscription_accessible_text_get_contents;
+  iface->get_contents_at = gtk_inscription_accessible_text_get_contents_at;
+  iface->get_caret_position = gtk_inscription_accessible_text_get_caret_position;
+  iface->get_selection = gtk_inscription_accessible_text_get_selection;
+  iface->get_attributes = gtk_inscription_accessible_text_get_attributes;
+  iface->get_default_attributes = gtk_inscription_accessible_text_get_default_attributes;
+  iface->get_extents = gtk_inscription_accessible_text_get_extents;
+  iface->get_offset = gtk_inscription_accessible_text_get_offset;
+}
+
+/* }}} */
+
+/* vim:set foldmethod=marker expandtab: */
